/*
 *   tappipe       TAP device <--> data stream converter
 *   Copyright (c) 2006 Michal Trojnara <Michal.Trojnara@mirt.net>
 *                 All Rights Reserved
 *
 *   Version:      1.02             (tappipe.c)
 *   Date:         2006.03.03
 *
 *   Author:       Michal Trojnara  <Michal.Trojnara@mirt.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <syslog.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <netinet/in.h>
#include <linux/if_tun.h>
#include <pthread.h>
#include <signal.h>

/**************************************** compile-time parameters */

/* max packet size */
#define MTU 2048

/* drop retrasmission within the timeout (in seconds) */
/* can be as high as 2MSL = 120 seconds */
#define TIMEOUT 5

/* size of drop table hash bucket */
/* memory usage is (8*BUCKET+4)*65536=2.25MB */
#define BUCKET 4

/* allowed retransmissions per second */
#define THRESHOLD 5

/**************************************** data structures */

#define LEN sizeof(u_long)

typedef struct {
    u_long len;
    u_char buf[MTU];
} PKT;

typedef union {
    PKT pkt;
    u_char buf[sizeof(PKT)];
} BUF;

typedef struct {
    u_long ip;
    u_long timestamp;
} SRC;

typedef struct {
    SRC src[BUCKET];
    u_long cycle;
} DROP_TABLE[65536];

/**************************************** global variables */

volatile static u_long timestamp=0; /* internal timer */
static pthread_t log_thread;        /* log thread to join at exit */
static DROP_TABLE drop_table;       /* retransmission hash table */

/* file descriptors */
static int signal_pipe[2];
volatile static int log_fd=-1, tap_fd=-1;

/* statistic data */
volatile static int pkt_in=0, pkt_out=0, bytes_in=0, bytes_out=0;
volatile static int retr_cnt=0, retr_tcp=0, retr_passed=0, retr_dropped=0;

/**************************************** logging */

static void *log_proc(void *param) {
    char buf[1024], *nl;
    int len, num;

    openlog("tappipe", LOG_CONS, LOG_DAEMON);
    len=0;
    while((num=read(log_fd, buf+len, sizeof(buf)-len-1))>0) {
        buf[len+num]='\0'; /* build a null-terminated string */
        len=strlen(buf); /* in case '\0' is logged */
        while((nl=index(buf, '\n'))!=NULL) { /* a new line */
            *nl='\0';
            syslog(LOG_ERR, "%s", buf);
            len-=nl-buf+1;
            memmove(buf, nl+1, len);
        }
    }
    return NULL;
}

static void log_exit(void) {
    fflush(stderr); /* just in case buffers are not empty */
    close(2); /* send EOF to the pipe */
    pthread_join(log_thread, (void **)NULL);
}

static void log_init(void) {
    int log_pipe[2];
    sigset_t newmask, oldmask;

    pipe(log_pipe);
    log_fd=log_pipe[0]; /* setup global variable */
    fcntl(log_fd, F_SETFD, FD_CLOEXEC);
    dup2(log_pipe[1], 2); /* dup 2 stderr */
    close(log_pipe[1]); /* no longer needed */

    sigfillset(&newmask);
    pthread_sigmask(SIG_SETMASK, &newmask, &oldmask); /* block signals */
    pthread_create(&log_thread, NULL, log_proc, NULL);
    pthread_sigmask(SIG_SETMASK, &oldmask, NULL); /* restore the mask */
    atexit(log_exit);
}

/**************************************** signals */

static void signal_handler(int sig) {
    write(signal_pipe[1], &sig, sizeof(int));
}

static void signal_init() {
    pipe(signal_pipe);
    signal(SIGPIPE, SIG_IGN); /* avoid 'broken pipe' signal */
    signal(SIGALRM, signal_handler);
    signal(SIGUSR1, signal_handler);
    signal(SIGTERM, signal_handler);
    signal(SIGQUIT, signal_handler);
    signal(SIGINT, signal_handler);
    signal(SIGHUP, signal_handler);
}

static void signal_process() {
    int sig;

    alarm(1);
    while(1) {
        while(read(signal_pipe[0], &sig, sizeof(int))!=sizeof(int))
            ; /* do nothing */
        switch(sig) {
        case SIGALRM:
            retr_cnt=0;
            ++timestamp;
            alarm(1);
            break;
        case SIGUSR1:
            fprintf(stderr,
                "in: %d packet(s) %d byte(s), out: %d packet(s), %d byte(s)\n",
                pkt_in, bytes_in, pkt_out, bytes_out);
            fprintf(stderr,
                "TCP packets: %d, retransmissions: %d passed, %d dropped\n",
                retr_tcp, retr_passed, retr_dropped);
            break;
        default:
            fprintf(stderr, "Received signal %d, terminating\n", sig);
            exit(0);
        }
    }
}

/**************************************** TAP */

#ifdef DEBUG
static void tap_debug(const u_char *buf, const int len) {
    int i, j;
    char txt[301];
    const char *hex="0123456789ABCDEF";

    i=0;
    j=0;
    do {
        txt[i++]=hex[(buf[j]&0xf0)>>4];
        txt[i++]=hex[buf[j]&0xf];
        j++;
        if(~j%2)
            txt[i++]=j%8 ? ' ' : '-';
    } while(j<len && i<sizeof(txt));
    txt[--i]='\0';
    syslog(LOG_NOTICE, "%d bytes: %s", len, txt);
}
#endif

/* avoid TCP-over-TCP meltdown */
static int tap_drop(const u_char *buf, const int len) {
    u_long id, i, cycle;
    SRC *src;

    if(len<34) /* to short to ba a valid IP packet (14B eth + 20B IP) */
        return 0; /* pass */
    if(buf[12]!=8 || buf[13]!=0) /* not ethernet type IPv4 */
        return 0; /* pass */
    if(buf[23]!=6) /* not TCP protocol */
        return 0; /* pass */
    ++retr_tcp;
    id=buf[18]+(buf[19]<<8); /* IP packet identification */
    for(i=0; i<BUCKET; i++) { /* cache bucket lookup */
        src=&drop_table[id].src[i];
        if(timestamp-src->timestamp<TIMEOUT && !memcmp(buf+26, &src->ip, 4))
            break; /* found a matching TCP segment */
    }
    if(i==BUCKET) { /* not found -> add one to the cyclic cache buffer */
        cycle=drop_table[id].cycle;
        src=&drop_table[id].src[cycle];
        memcpy(&src->ip, buf+26, 4);
        src->timestamp=timestamp;
        drop_table[id].cycle=(cycle+1)%BUCKET;
        return 0; /* pass */
    }
    if(retr_cnt++<THRESHOLD) { /* within the maximum retransmission rate ? */
        ++retr_passed;
        return 0; /* pass */
    }
    ++retr_dropped;
#ifdef DEBUG
    tap_debug(buf, len);
#endif
    return 1; /* drop */
}

static void *tap_read(void *param) {
    BUF buf;
    int num;

    while(1) {
        num=read(tap_fd, buf.pkt.buf, MTU);
        switch(num) {
        case -1:
            perror("Reading TAP");
            exit(1);
        case 0:
            fprintf(stderr, "TAP closed\n");
            exit(1);
        }
        if(tap_drop(buf.pkt.buf, num))
            continue;
        bytes_out+=num; /* sent to pipe */
        buf.pkt.len=htonl(num);
        num=write(1, buf.buf, LEN+num);
        if(num<0) {
            perror("Writing PIPE");
            exit(1);
        }
        ++pkt_out; /* sent to pipe */
    }
    return NULL;
}

static void *tap_write(void *param) {
    BUF buf;
    int len, num;

    len=0;
    while(1) {
        num=read(0, buf.buf+len, sizeof(buf)-len);
        switch(num) {
        case -1:
            perror("Reading PIPE");
            exit(1);
        case 0:
            fprintf(stderr, "PIPE closed\n");
            exit(1);
        }
        len+=num;
        while(len>=LEN && len>=LEN+ntohl(buf.pkt.len)) {
            num=write(tap_fd, buf.pkt.buf, ntohl(buf.pkt.len));
            if(num<0) {
                perror("Writing TAP");
                exit(1);
            }
            bytes_in+=num; /* read from pipe */
            num=LEN+ntohl(buf.pkt.len);
            len-=num;
            memmove(buf.buf, buf.buf+num, len);
            ++pkt_in; /* read from pipe */
        }
    }
    return NULL;
}

static void tap_init(char *script) {
    struct ifreq ifr;
    char cmd[256];
    int tmpfd, err;
    pthread_t tap_thread_read, tap_thread_write;
    sigset_t newmask, oldmask;

    tap_fd=open("/dev/net/tun", O_RDWR);
    if(tap_fd<0) {
        perror("open: /dev/net/tun");
        exit(1);
    }
    fcntl(tap_fd, F_SETFD, FD_CLOEXEC);
    memset(&ifr, 0, sizeof(ifr));
    ifr.ifr_flags = IFF_TAP | IFF_NO_PI;
    if(ioctl(tap_fd, TUNSETIFF, (void *)&ifr)<0) {
        perror("ioctl: /dev/net/tun");
        exit(1);
    }
    snprintf(cmd, sizeof(cmd), "%s %s\n", script, ifr.ifr_name);
    tmpfd=dup(1); /* save copy of stdout descriptor */
    fcntl(tmpfd, F_SETFD, FD_CLOEXEC);
    dup2(2, 1); /* redirect stdout to stderr */
    err=system(cmd);
    dup2(tmpfd, 1); /* restore stdout */
    close(tmpfd); /* close temporary descriptor */
    if(err) {
        fprintf(stderr, "system() failed\n");
        exit(1);
    }
    memset(&drop_table, 0, sizeof(drop_table));
    sigfillset(&newmask);
    pthread_sigmask(SIG_SETMASK, &newmask, &oldmask); /* block signals */
    pthread_create(&tap_thread_read, NULL, tap_read, NULL);
    pthread_create(&tap_thread_write, NULL, tap_write, NULL);
    pthread_sigmask(SIG_SETMASK, &oldmask, NULL); /* restore the mask */
}

/**************************************** main() function */

int main(int argc, char *argv[]) {
    signal_init();
    log_init();
    if(argc<2) {
        fprintf(stderr, "Interface configuration script not specified\n");
        exit(1);
    }
    tap_init(argv[1]);
    fprintf(stderr, "tappipe started\n");
    signal_process();
    return 0;
}

/**************************************** end of tappipe.c file */
